from inspect import isclass
from typing import Union, List, Type

from janis_core.utils.scatter import ScatterDescription, ScatterMethods

from janis_core.types import Array

from janis_core.workflow.workflow import Tool, Workflow


class BatchRunTool(Workflow):
    def __init__(
        self,
        innertool: Union[Tool, Type[Tool]],
        fields: List[str],
        group_by: str = None,
    ):
        self.inner: Tool = innertool() if isclass(innertool) else innertool
        self.fields = set(fields)
        self.group_by = group_by

        super().__init__()

    def friendly_name(self):
        return "Multiplied " + self.inner.friendly_name()

    def constructor(self):
        ins = self.inner.tool_inputs()

        inkeys = set(i.id() for i in ins)
        invalid_keys = self.fields - inkeys
        if len(invalid_keys) > 0:
            raise Exception(
                f"Couldn't create multitool from fields {', '.join(invalid_keys)} "
                f"as they do not exist on '{self.inner.id()}'"
            )

        innode_map = {}

        for i in ins:
            intype = i.intype
            default = i.default
            if i.id() in self.fields:
                intype = Array(intype)
                default = None
            innode_map[i.id()] = self.input(i.id(), intype, default=default, doc=i.doc)

        self.step(
            self.inner.id(),
            self.inner(**innode_map),
            scatter=ScatterDescription(list(self.fields), ScatterMethods.dot),
        )

        if isinstance(self.inner, Workflow):
            # We can do special output_folder grouping
            for oid, o in self.output_nodes.items():
                folders = o.output_folder
                if self.group_by:
                    if not folders:
                        folders = []
                    folders.append(self[self.group_by])

                self.output(
                    oid, o.datatype, output_folder=folders, output_name=o.output_name
                )
        else:
            for o in self.inner.tool_outputs():
                f = self[self.group_by] if o else None
                self.output(o.id(), o.outtype, output_folder=f)

    def id(self) -> str:
        return "multiple_" + self.inner.id()



"""
    This file declares the 'entry_points' that Janis will look for in its operations, and
    that your tool can bind to in order to provide additional functionality to Janis.

    > Check this awesome tutorial about Snek Inc for more information about entry-points:
        https://amir.rachum.com/blog/2017/07/28/python-entry-points/

    This is how all of janis works, the unix and bioinformatics tools are registered under
    all three of 'janis.extension', 'janis.types' and 'janis.tools'. janis-runner is just
    bound under 'janis.extension' to provide 'janis.runner' and CLI functionality.
"""


"""
    All extensions registered under this entrypoint will be added as an export to 
    janis.{yourextension}. Additionally, they will be added to the python sys.path 
    so you can just call'import janis.{yourextension}'.
"""
EXTENSIONS = "janis.extension"


"""
    All extensions that want to declare types should register a 'janis.types' extension.
    The shed (`janis.shed`) will look up this entry point and traverse any child modules
    looking for data types. 
    
    Note: Users can still access your type through a direct Python import (preferred)

"""
DATATYPES = "janis.types"

"""
    All extensions that want to provide tools should register a 'janis.tools' extension.
    The shed (`janis.shed`) will look up this entry point and traverse any child modules
    looking for tools. 
    
    Note: Users can still access your tool through a direct Python import (preferred)
"""
TOOLS = "janis.tools"

"""
    Polished pipelines should be extended under the janis.pipelines class. This will allow
    the documentation to produce a separate page called "Pipelines" with these workflows.
    They are also rendered slightly differently.
"""
PIPELINES = "janis.pipelines"


"""
    Templates used by the janis-assistant for running can be extended by providing one
    additional template per entry point. The name of entry point will be used as the
    corresponding key, and these WILL override internal templates.
"""
TEMPLATES = "janis.templates"
